package es.weso.shex.report

import org.scalatest.{EitherValues, FunSpec, Matchers}
import com.typesafe.config._
import java.io._
import java.nio.file.Paths
import es.weso.shex.manifest.ManifestPrefixes._
import es.weso.rdf.RDFReader
import es.weso.rdf.jena.RDFAsJenaModel
import es.weso.rdf.nodes.{BNode, IRI}
import es.weso.rdf.parser.RDFParser
import es.weso.shapeMaps.{BNodeLabel => BNodeMapLabel, IRILabel => IRIMapLabel, Start => StartMap, _}
import es.weso.shex.validator.{ExternalIRIResolver, Validator}
import es.weso.shex._
import es.weso.shex.manifest.JsonResult
import io.circe.syntax._
import es.weso.utils.UriUtils._
import scala.collection.mutable

// TODO: Remove duplication between ValidationTest and ValidationFailureTest

class ValidationFromManifestCompatTest extends FunSpec with Matchers with RDFParser with EitherValues {

  // If the following variable is None, it runs all tests
  // Otherwise, it runs only the test whose name is equal to the value of this variable
  val nameIfSingle: Option[String] =
    None
    // Some("open3Onedotclosecard23_pass-p1p2")

  val counter = Counter()
  val conf: Config = ConfigFactory.load()
  val manifestFile = new File(conf.getString("manifestFile"))
  val outFile = conf.getString("EarlReportFile")
  val baseIRI: Option[IRI] = Some(IRI("https://raw.githubusercontent.com/shexSpec/shexTest/master/validation/manifest"))

  describe("Generate W3c EARL report") {
    RDFAsJenaModel.fromFile(manifestFile, "TURTLE", baseIRI) match {
      case Left(e) => info(s"Error reading manifestTest file: $e")
      case Right(rdf) => {
        val report = prepareReport(rdf)
        val earlModel = report.generateEARL
        earlModel.write(new FileOutputStream(outFile), "TURTLE")
        //          info(s"Report written to $outFile")
        /*for (item <- report.items) {
            it(s"Should pass ${item.name}") {
              if (!item.passed) {
                fail(s"Failed: ${item.testType}: ${item.moreInfo}")
              } else {
                info(s"OK: ${item.name}")
              }
            }
           }*/
      }
    }
  }

  def prepareReport(manifestRdf: RDFReader): Report = {
    val base = Paths.get(".").toUri
    val report = Report.empty


    // Validation tests
    for (triple <- manifestRdf.triplesWithType(sht_ValidationTest).right.value) {
      val node = triple.subj
      val nodeStr = node.getLexicalForm
      val triples = manifestRdf.triplesWithSubjectPredicate(node, mf_name).right.value
      val name = triples.map(_.obj).head.getLexicalForm
      if (nameIfSingle == None || nameIfSingle.getOrElse("") === name) {
        it(s"validationTest: $name") {
          val tryReport = for {
            name      <- stringFromPredicate(mf_name)(node, manifestRdf)
            action    <- objectFromPredicate(mf_action)(node, manifestRdf)
            traits     <- objectsFromPredicate(sht_trait)(node, manifestRdf)
            maybeResult  <- iriFromPredicateOptional(mf_result)(node, manifestRdf)
            schemaIRI <- iriFromPredicate(sht_schema)(action, manifestRdf)
            resolvedSchema = base.resolve(schemaIRI.uri)
            schemaStr <- derefUri(resolvedSchema)
            schema  <- Schema.fromString(schemaStr, "SHEXC", Some(schemaIRI)) // Some(resolvedSchema.toString))
//            _ <- { println(s"BaseIRI: $baseIRI\n$schemaIRI"); Right(())}
            dataIRI <- iriFromPredicate(sht_data)(action, manifestRdf)
            strData <- derefUri(dataIRI.uri)
            data           <- RDFAsJenaModel.fromChars(strData, "TURTLE", baseIRI)
            maybeFocus <- objectFromPredicateOptional(sht_focus)(action, manifestRdf)
            maybeMap  <- iriFromPredicateOptional(sht_map)(action, manifestRdf)
            maybeShape <- objectFromPredicateOptional(sht_shape)(action, manifestRdf)
            shapeExterns <- iriFromPredicateOptional(sht_shapeExterns)(action, manifestRdf)
            shapeExternsResolved = shapeExterns.map(iri => IRI(base.resolve(iri.uri)))
            lbl = maybeShape match {
              case None           => StartMap: ShapeMapLabel
              case Some(i: IRI)   => IRIMapLabel(i)
              case Some(b: BNode) => BNodeMapLabel(b)
              case Some(other) => {
                IRIMapLabel(IRI(s"UnknownLabel"))
              }
            }
            ok <- if (traits contains sht_Greedy) {
              counter.add(s"Greedy: $name")
              Right(s"Greedy")
            } else maybeFocus match {
              case Some(focus) => {
                val shapeMap = FixedShapeMap(Map(focus -> Map(lbl -> Info())), data.getPrefixMap, schema.prefixMap)
                for {
                  resultShapeMap <- Validator(schema, ExternalIRIResolver(shapeExternsResolved)).validateShapeMap(data, shapeMap).toEitherS
                  ok <- if (resultShapeMap.getConformantShapes(focus) contains lbl)
                        Right(s"Focus $focus conforms to $lbl")
                else Left(s"Focus $focus does not conform to shape $lbl\nResultMap:\n$resultShapeMap" ++
                  s"\nData: \n${strData}\nSchema: ${schemaStr}\n" ++
                  s"${resultShapeMap.getInfo(focus,lbl)}\n" ++
                  s"Schema: ${schema}\n" ++
                  s"Data: ${data}"
                )
              } yield ok
              }
              case None => (maybeMap,maybeResult) match {
                case (Some(smapIRI), Some(resultIRI)) => {
                  val r: Either[String,String] = for {
                    smapStr <- derefUri(smapIRI.uri)
                    resultMapStr <- derefUri(resultIRI.uri)
                    sm <- ShapeMap.fromJson(smapStr)
                    fixedShapeMap <- ShapeMap.fixShapeMap(sm, manifestRdf, manifestRdf.getPrefixMap(),schema.prefixMap)
                    resultShapeMap <- Validator(schema).validateShapeMap(data, fixedShapeMap).toEitherS
                    jsonResult <- JsonResult.fromJsonString(resultMapStr)
                    result <- if (jsonResult.compare(resultShapeMap)) Right(s"Json results match resultShapeMap")
                    else Left(s"Json results are different. Expected: ${jsonResult.asJson.spaces2}\nObtained: ${resultShapeMap.toString}")
                  } yield result
                  r
                }
                case other => {
                  Left(s"No focus and no map/result: $other")
                }
              }
            }
          } yield ok

          val testReport = tryReport match {
            case Right(msg) => {
              SingleTestReport(passed = true,
                               name = name,
                               uriTest = nodeStr,
                               testType = sht_ValidationTest.str,
                               moreInfo = s"$msg")
            }
            case Left(e) => {
              SingleTestReport(passed = false,
                               name = name,
                               uriTest = node.getLexicalForm,
                               testType = sht_ValidationTest.str,
                               moreInfo = s"Error:${e}")
            }
          }
          report.addTestReport(testReport)
          tryReport.fold(e => fail(s"Error: $e"), msg => info(s"$msg"))
        }
      }
    }

    // Validation tests
    for (triple <- manifestRdf.triplesWithType(sht_ValidationFailure).right.value) {
      val node    = triple.subj
      val nodeStr = node.getLexicalForm
      val triples = manifestRdf.triplesWithSubjectPredicate(node, mf_name).right.value
      val name = triples.map(_.obj).head.getLexicalForm
      if (nameIfSingle == None || nameIfSingle.getOrElse("") === name) {
        it(s"ValidationFailureTest: $name") {
        val tryReport = for {
          name      <- stringFromPredicate(mf_name)(node, manifestRdf)
          traits     <- objectsFromPredicate(sht_trait)(node, manifestRdf)
          maybeResult  <- iriFromPredicateOptional(mf_result)(node, manifestRdf)
          action    <- objectFromPredicate(mf_action)(node, manifestRdf)
          schemaIRI <- iriFromPredicate(sht_schema)(action, manifestRdf)
          schemaStr <- derefUri(schemaIRI.uri)
          schema  <- Schema.fromString(schemaStr, "SHEXC", Some(schemaIRI))
          dataIRI <- iriFromPredicate(sht_data)(action, manifestRdf)
          strData <- derefUri(dataIRI.uri)
          data           <- RDFAsJenaModel.fromChars(strData, "TURTLE", baseIRI)
          maybeFocus          <- objectFromPredicateOptional(sht_focus)(action, manifestRdf)
          maybeMap  <- iriFromPredicateOptional(sht_map)(action, manifestRdf)
          maybeShape     <- objectFromPredicateOptional(sht_shape)(action, manifestRdf)
          shapeExterns <- iriFromPredicateOptional(sht_shapeExterns)(action, manifestRdf)
          shapeExternsResolved = shapeExterns.map(iri => IRI(base.resolve(iri.uri)))
          lbl = maybeShape match {
            case None           => StartMap: ShapeMapLabel
            case Some(i: IRI)   => IRIMapLabel(i)
            case Some(b: BNode) => BNodeMapLabel(b)
            case Some(other) => {
              IRIMapLabel(IRI(s"UnknownLabel"))
            }
          }
          ok <- if (traits contains sht_Greedy) {
            counter.add(s"Greedy: $name")
            Right(s"Greedy")
          } else maybeFocus match {
            case Some(focus) => {
              val shapeMap = FixedShapeMap(Map(focus -> Map(lbl -> Info())), data.getPrefixMap, schema.prefixMap)
              for {
                resultShapeMap <- Validator(schema,ExternalIRIResolver(shapeExternsResolved)).validateShapeMap(data, shapeMap).toEitherS
                ok <- if (resultShapeMap.getNonConformantShapes(focus) contains lbl)
                       Right(s"Focus $focus does not conforms to $lbl as expected")
                else
                Left(s"Focus $focus does conform to shape $lbl and should not\nResultMap:\n$resultShapeMap" ++
                  s"\nData: \n${strData}\nSchema: ${schemaStr}\n" ++
                  s"${resultShapeMap.getInfo(focus,lbl)}\n" ++
                  s"Schema: ${schema}\n" ++
                  s"Data: ${data}"
                )
              } yield ok
             }
            case None => (maybeMap,maybeResult)  match {
              case (Some(smapIRI), Some(resultIRI)) => {
                val r: Either[String,String] = for {
                 smapStr <- derefUri(smapIRI.uri)
                 resultMapStr <- derefUri(resultIRI.uri)
                 sm <- ShapeMap.fromJson(smapStr)
                 fixedShapeMap <- ShapeMap.fixShapeMap(sm, manifestRdf,manifestRdf.getPrefixMap(),schema.prefixMap)
                 resultShapeMap <- Validator(schema).validateShapeMap(data, fixedShapeMap).toEitherS
                 jsonResult <- JsonResult.fromJsonString(resultMapStr)
                 result <- if (jsonResult.compare(resultShapeMap)) Right(s"Json results match resultShapeMap")
                 else Left(s"Json results are different. Expected: ${jsonResult.asJson.spaces2}\nObtained: ${resultShapeMap.toString}")
                } yield result
                r
              }
              case other => {
                Left(s"No focus and no map/result: $other!")
              }
            }
          }
        } yield ok

        val testReport = tryReport match {
          case Right(msg) => {
              SingleTestReport(passed = true,
                               name = name,
                               uriTest = nodeStr,
                               testType = sht_ValidationTest.str,
                               moreInfo = s"Failed as expected with error: ${msg}")
          }
          case Left(e) => {
            SingleTestReport(passed = false,
                             name = name,
                             uriTest = nodeStr,
                             testType = sht_ValidationTest.str,
                             moreInfo = s"Error ${e}")
          }
        }
        report.addTestReport(testReport)
        tryReport.fold(e => fail(s"Error: $e"), msg => info(s"$msg"))
      }
     }
    }

    it(s"Inform counter") {
      info(s"Counter:\n${counter.toString}")
    }
    report
  }


}

case class Counter() {
  private val msgs: collection.mutable.Buffer[String] = mutable.Buffer()
  def add(msg: String): Unit = {
    msgs += msg
  }

  override def toString(): String = {
    msgs.mkString("\n") ++ s"\nTotal: ${msgs.size}"
  }


}
